#include <lexer/Lexer.h>

#include <cctype>

namespace cli {

Lexer::Token Lexer::nextToken ( bool readNextLine ) {
	Token res { "", "", TokenType::ERROR };

	switch ( m_hint ) {
	case Hint::NONE:
		goto q0;
	case Hint::FILE:
		goto qFile;
	case Hint::TYPE:
		goto qType;
	}

q0:	if ( m_source.getCharacter ( ) == EOF ) {
		res.m_type = TokenType::EOT;
		return res;
	}
	if ( m_source.getCharacter ( ) == '\0' ) {
		res.m_type = TokenType::EOS;
		return res;
	}
	if ( isspace ( m_source.getCharacter ( ) ) ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		goto q0;
	}
	if ( m_source.getCharacter ( ) == '<' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		res.m_type = TokenType::IN_REDIRECT;
		return res;
	}
	if ( m_source.getCharacter ( ) == '>' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		res.m_type = TokenType::OUT_REDIRECT;
		return res;
	}
	if ( m_source.getCharacter ( ) == '(' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		res.m_type = TokenType::LEFT_PAREN;
		return res;
	}
	if ( m_source.getCharacter ( ) == ')' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		res.m_type = TokenType::RIGHT_PAREN;
		return res;
	}
	if ( m_source.getCharacter ( ) == '{' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		res.m_type = TokenType::LEFT_BRACE;
		return res;
	}
	if ( m_source.getCharacter ( ) == '}' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		res.m_type = TokenType::RIGHT_BRACE;
		return res;
	}
	if ( m_source.getCharacter ( ) == '[' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		res.m_type = TokenType::LEFT_BRACKET;
		return res;
	}
	if ( m_source.getCharacter ( ) == ']' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		res.m_type = TokenType::RIGHT_BRACKET;
		return res;
	}
	if ( m_source.getCharacter ( ) == '@' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		res.m_type = TokenType::AT_SIGN;
		return res;
	}
	if ( m_source.getCharacter ( ) == '$' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		res.m_type = TokenType::DOLAR_SIGN;
		return res;
	}
	if ( m_source.getCharacter ( ) == '&' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		res.m_type = TokenType::AMPERSAND_SIGN;
		return res;
	}
	if ( m_source.getCharacter ( ) == '|' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		res.m_type = TokenType::PIPE_SIGN;
		return res;
	}
	if ( m_source.getCharacter ( ) == '^' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		res.m_type = TokenType::CARET_SIGN;
		return res;
	}
	if ( m_source.getCharacter ( ) == ':' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		res.m_type = TokenType::COLON_SIGN;
		return res;
	}
	if ( m_source.getCharacter ( ) == ';' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		res.m_type = TokenType::SEMICOLON_SIGN;
		return res;
	}
	if ( m_source.getCharacter ( ) == '=' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		res.m_type = TokenType::EQUAL_SIGN;
		return res;
	}
	if ( m_source.getCharacter ( ) == '#' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		res.m_type = TokenType::HASH_SIGN;
		return res;
	}
	if ( m_source.getCharacter ( ) == '-' ) {
		res.m_raw += m_source.getCharacter ( );
		res.m_value += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		goto q2;
	}

	if ( m_source.getCharacter ( ) == '"' ) {
		res.m_raw += m_source.getCharacter ( );
		res.m_type = TokenType::STRING;
		m_source.advance ( true );
		goto q4;
	}

	if ( ( m_source.getCharacter ( ) >= '0' && m_source.getCharacter ( ) <= '9' ) ) {
		res.m_raw += m_source.getCharacter ( );
		res.m_value += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		goto q1;
	}

	if ( ( m_source.getCharacter ( ) >= 'a' && m_source.getCharacter ( ) <= 'z' )
	  || ( m_source.getCharacter ( ) >= 'A' && m_source.getCharacter ( ) <= 'Z' ) ) {
		res.m_raw += m_source.getCharacter ( );
		res.m_value += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		goto q3;
	}

	if ( m_source.getCharacter ( ) == '\\' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( true );
		goto q3Escape;
	}

	res.m_type = TokenType::ERROR;
	return res;

q1:	if ( m_source.getCharacter ( ) == '\0' ) {
		res.m_type = TokenType::INTEGER;
		return res;
	}
	if ( ( m_source.getCharacter ( ) >= '0' && m_source.getCharacter ( ) <= '9' ) ) {
		res.m_raw += m_source.getCharacter ( );
		res.m_value += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		goto q1;
	} else {
		res.m_type = TokenType::INTEGER;
		return res;
	}

q2:	if ( m_source.getCharacter ( ) == '\0' ) {
		res.m_value = "";
		res.m_type = TokenType::DASH_SIGN;
		return res;
	}
	if ( ( m_source.getCharacter ( ) >= '0' && m_source.getCharacter ( ) <= '9' ) ) {
		res.m_raw += m_source.getCharacter ( );
		res.m_value += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		goto q1;
	}

	if ( ( m_source.getCharacter ( ) >= 'a' && m_source.getCharacter ( ) <= 'z' )
	  || ( m_source.getCharacter ( ) >= 'A' && m_source.getCharacter ( ) <= 'Z' )
	  || ( m_source.getCharacter ( ) == ':' ) ) {
		res.m_raw += m_source.getCharacter ( );
		res.m_value += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		goto q3;
	} else if ( m_source.getCharacter ( ) == '\\' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( true );
		goto q3Escape;
	}

	res.m_value = "";
	res.m_type = TokenType::DASH_SIGN;
	return res;

q3:	if ( m_source.getCharacter ( ) == '\0' ) {
		res.m_type = is_kw ( res.m_value );
		return res;
	}
	if ( ( m_source.getCharacter ( ) >= '0' && m_source.getCharacter ( ) <= '9' )
	  || ( m_source.getCharacter ( ) >= 'a' && m_source.getCharacter ( ) <= 'z' )
	  || ( m_source.getCharacter ( ) >= 'A' && m_source.getCharacter ( ) <= 'Z' )
	  || ( m_source.getCharacter ( ) == ':' ) ) {
		res.m_raw += m_source.getCharacter ( );
		res.m_value += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		goto q3;
	} else if ( m_source.getCharacter ( ) == '\\' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( true );
		goto q3Escape;
	} else {
		res.m_type = is_kw ( res.m_value );
		return res;
	}

q3Escape:
	if ( m_source.getCharacter ( ) == '\0' ) {
		res.m_type = TokenType::ERROR;
		return res;
	}

	res.m_raw += m_source.getCharacter ( );
	res.m_value += m_source.getCharacter ( );
	m_source.advance ( readNextLine );
	goto q3;

q4:	if ( m_source.getCharacter ( ) == '\0' ) {
		res.m_type = TokenType::ERROR;
		return res;
	}
	if ( m_source.getCharacter ( ) == '"' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		goto q6;
	}
	if ( m_source.getCharacter ( ) == '\\' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( true );
		goto q5;
	} else {
		res.m_raw += m_source.getCharacter ( );
		res.m_value += m_source.getCharacter ( );
		m_source.advance ( true );
		goto q4;
	}

q5:	if ( m_source.getCharacter ( ) == '\0' ) {
		res.m_type = TokenType::ERROR;
		return res;
	}

	if ( m_source.getCharacter ( ) == 'n' ) {
		res.m_raw += m_source.getCharacter ( );
		res.m_value += '\n';
	} else if ( m_source.getCharacter ( ) == 't' ) {
		res.m_raw += m_source.getCharacter ( );
		res.m_value += '\t';
	} else if ( m_source.getCharacter ( ) == '"' ) {
		res.m_raw += m_source.getCharacter ( );
		res.m_value += '"';
	} else {
		res.m_raw += m_source.getCharacter ( );
		res.m_value += m_source.getCharacter ( );
	}

	m_source.advance ( true );
	goto q4;

q6:	if ( m_source.getCharacter ( ) == '\0' ) {
		return res;
	}
	if ( isspace ( m_source.getCharacter ( ) ) ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		goto q6;
	}
	if ( m_source.getCharacter ( ) == '"' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( true );
		goto q4;
	} else {
		return res;
	}

qFile:
	if ( m_source.getCharacter ( ) == EOF ) {
		res.m_type = TokenType::EOT;
		return res;
	}
	if ( m_source.getCharacter ( ) == '\0' ) {
		res.m_type = TokenType::EOS;
		return res;
	}
	if ( isspace ( m_source.getCharacter ( ) ) ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		goto qFile;
	}
	if ( ( m_source.getCharacter ( ) >= '0' && m_source.getCharacter ( ) <= '9' )
	  || ( m_source.getCharacter ( ) >= 'a' && m_source.getCharacter ( ) <= 'z' )
	  || ( m_source.getCharacter ( ) >= 'A' && m_source.getCharacter ( ) <= 'Z' )
	  ||   m_source.getCharacter ( ) == '/' || m_source.getCharacter ( ) == '.' || m_source.getCharacter ( ) == '-'
	  ||   m_source.getCharacter ( ) == '~' || m_source.getCharacter ( ) == '_' || m_source.getCharacter ( ) == ':' ) {
		res.m_raw += m_source.getCharacter ( );
		res.m_value += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		goto qFile2;
	} else {
		goto q0;
	}

qFile2:
	if ( m_source.getCharacter ( ) == '\0' ) {
		res.m_type = TokenType::FILE;
		return res;
	}
	if ( ( m_source.getCharacter ( ) >= '0' && m_source.getCharacter ( ) <= '9' )
	  || ( m_source.getCharacter ( ) >= 'a' && m_source.getCharacter ( ) <= 'z' )
	  || ( m_source.getCharacter ( ) >= 'A' && m_source.getCharacter ( ) <= 'Z' )
	  ||   m_source.getCharacter ( ) == '/' || m_source.getCharacter ( ) == '.' || m_source.getCharacter ( ) == '-'
	  ||   m_source.getCharacter ( ) == '~' || m_source.getCharacter ( ) == '_' || m_source.getCharacter ( ) == ':' ) {
		res.m_raw += m_source.getCharacter ( );
		res.m_value += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		goto qFile2;
	} else if ( m_source.getCharacter ( ) == '\\' ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( true );
		goto qFileEscape;
	} else {
		res.m_type = TokenType::FILE;
		return res;
	}

qFileEscape:
	if ( m_source.getCharacter ( ) == '\0' ) {
		res.m_type = TokenType::ERROR;
		return res;
	}

	res.m_raw += m_source.getCharacter ( );
	res.m_value += m_source.getCharacter ( );
	m_source.advance ( readNextLine );
	goto qFile2;

qType:
	if ( m_source.getCharacter ( ) == EOF ) {
		res.m_type = TokenType::EOT;
		return res;
	}
	if ( m_source.getCharacter ( ) == '\0' ) {
		res.m_type = TokenType::EOS;
		return res;
	}
	if ( isspace ( m_source.getCharacter ( ) ) ) {
		res.m_raw += m_source.getCharacter ( );
		m_source.advance ( readNextLine );
		goto qType;
	}
	if ( m_source.getCharacter ( ) == ')' ) {
		goto q0;
	}

	{
		unsigned lparens = 0;
		while ( m_source.getCharacter ( ) != '\0' ) {
			if ( m_source.getCharacter ( ) == '(' )
				++ lparens;
			else if ( m_source.getCharacter ( ) == ')' && lparens > 0 )
				-- lparens;
			else if ( m_source.getCharacter ( ) == ')' ) {
				break;
			}
			res.m_raw += m_source.getCharacter ( );
			res.m_value += m_source.getCharacter ( );
			m_source.advance ( readNextLine );
		}
		res.m_value = ext::rtrim ( res.m_value );

		res.m_type = TokenType::TYPE;
		return res;
	}
}

} /* namespace cli */
